#coding:utf-8

from abc import ABCMeta, abstractmethod

from sqlalchemy.orm import load_only
from sqlalchemy.sql.functions import min
from sqlalchemy import and_, or_

from m3.actions.exceptions import ApplicationLogicException

from kinder.core.dict.models import UnitKind
from kinder.core.group.models import GroupStatusEnum
from kinder.core.declaration_status.models import DSS

from gisdo.constants import (get_report_period, AGE_CATEGORY_DELTAS,
                             GOVERMENT, ALL)
from alchemy_models import (AlchemyDeclaration, AlchemyDeclarationUnit,
                            AlchemyUnit, AlchemyChildren, AlchemyWorkType,
                            AlchemyDeclarationStatus, AlchemyGroup, AlchemyPupil,
                            AlchemyGroupType, AlchemyGroupWorkType)


class BaseIndex(object):
    """
    Базовый класс для показателей
    """

    __metaclass__ = ABCMeta

    def __init__(self, unit, session):

        self.unit = unit
        self.unit_mo_id = unit.get_mo().id
        self.dou_type = unit.dou_type

        self.session = session

    @classmethod
    def get_algorithm(cls, index_type):

        try:
            clazz = cls.ALGORITHM_MAP[index_type]
        except KeyError:
            raise ApplicationLogicException(u'Такой показатель не реализован')

        return clazz

    @abstractmethod
    def get_count(self):
        pass


class BaseAlgorithm(object):

    __metaclass__ = ABCMeta

    def __init__(self, session, unit):

        self.unit = unit
        self.unit_mo_id = unit.get_mo().id
        self.session = session

    def _kids(self, **kwargs):
        """
        Дети с датой рождения из age_range
        """

        age_range = kwargs.get('age_range', ALL)
        down, up = AGE_CATEGORY_DELTAS[age_range]
        kids = (self.session.query(AlchemyChildren).filter(AlchemyChildren.date_of_birth >= down,
                                                           AlchemyChildren.date_of_birth < up).options(load_only('id'))
                                                   .subquery())
        return kids

    @abstractmethod
    def get_result_set(self):
        pass


class BaseApplicationAlgorithm(BaseAlgorithm):

    __metaclass__ = ABCMeta

    def _dou_units(self):
        """
        Достаём для каждого желаемого учреждения привязанное к нему учреждение
        """
        # Учреждение с типом ДОУ
        dou_units = (self.session.query(AlchemyUnit).filter(AlchemyUnit.kind_id == UnitKind.DOU)
                                                    .options(load_only('id'))
                                                    .subquery())
        # JOIN с желаемыми учреждениями.
        return (self.session.query(AlchemyDeclarationUnit, 'unit_id').join(dou_units)
                                                                     .options(load_only('id', 'ord', 'declaration_id'))
                                                                     .subquery())

    def _get_base_query(self, **kwargs):
        """
        Базовый запрос для группы показателей "Общее количество поданных заявлений..."
        """

        # Фильтрация по отчетному периоду.
        begin_report_period, end_report_period = get_report_period()

        _base_query = (self.session.query(AlchemyDeclaration, 'ord', 'unit_id')
                                   .filter(AlchemyDeclaration.date >= begin_report_period,
                                           AlchemyDeclaration.date <= end_report_period,
                                           AlchemyDeclaration.mo_id == self.unit_mo_id)
                                   .join(self._dou_units()).join(self._kids(**kwargs))
                                   .options(load_only('portal', 'status_id', 'work_type_id', 'id'))
                                   .subquery())
        return _base_query

    @abstractmethod
    def get_result_set(self, **kwargs):
        pass


class ApplicationCountByAge(BaseApplicationAlgorithm):

    def get_result_set(self, **kwargs):
        """
        Количество заявлений по возрастам
        """

        return self._get_base_query(**kwargs)


class ApplicationCountByDelivery(BaseApplicationAlgorithm):

    def get_result_set(self, **kwargs):
        """
        По типу подачи: с портала или при личном обращении
        """

        portal = kwargs.get('portal', 0)
        base = self._get_base_query()

        return self.session.query(base).filter(base.c.portal == portal).subquery()


class ApplicationCountByGroupType(BaseApplicationAlgorithm):

    def get_result_set(self, **kwargs):
        """
        Группы кратковременного пребывания
        """

        base = self._get_base_query(**kwargs)
        return self.session.query(base).join(AlchemyWorkType).filter(AlchemyWorkType.code == '3-5').subquery()


class ApplicationCountByDenied(BaseApplicationAlgorithm):

    def get_result_set(self, **kwargs):
        """
        Количество отказов.
        """

        base = self._get_base_query(**kwargs)
        return (self.session.query(base).join(AlchemyDeclarationStatus)
                                        .filter(or_(AlchemyDeclarationStatus.code == DSS.DIDNT_COME,
                                                    AlchemyDeclarationStatus.code == DSS.REFUSED))
                                        .subquery())


class ApplicationIndex(BaseIndex):
    """
    Запросы группы "Общее количество поданных заявлений"
    """

    COUNT_BY_AGE = 1
    COUNT_BY_DELIVERY = 2
    COUNT_BY_GROUP_TYPE = 3
    COUNT_DENIED = 4

    ALGORITHM_MAP = {
        COUNT_BY_AGE: ApplicationCountByAge,
        COUNT_BY_DELIVERY: ApplicationCountByDelivery,
        COUNT_BY_GROUP_TYPE: ApplicationCountByGroupType,
        COUNT_DENIED: ApplicationCountByDenied
    }

    def get_count(self, dou_type=GOVERMENT, index_type=COUNT_BY_AGE, **kwargs):
        """
        Для всех показателей есть общий тип запроса. В конкретных показателях навешиваются дополнительные фильтры.
        В зависимости от параметра index_type управление пробрасывается конкретной функции для добавления специфичных
        для показателя фильтров.
        """

        # Если необходимо посчитать количество заявлений в негосударственное или частное ДОУ для государственного сада,
        # то сразу возвращаем 0.
        if (not self.dou_type and dou_type != GOVERMENT) or (self.dou_type and self.dou_type.code != dou_type):
            return 0

        result_set = (ApplicationIndex.get_algorithm(index_type)(session=self.session, unit=self.unit)
                                      .get_result_set(**kwargs))

        # base - это таблица с тремя столбцами
        # id | unit_id | ord
        # Нам необходимо для каждой заявки отобрать самое приоритетное учреждение Unit.
        # SELECT * FROM base
        # JOIN (SELECT id, min(ord) as min_ord FROM base) as T
        # ON T.id = base.id AND T.min_ord = base.ord
        result_set_min_ord = (self.session.query(result_set.c.id, min(result_set.c.ord).label('min_ord'))
                                          .group_by(result_set.c.id).subquery())
        return (self.session.query(result_set).join(result_set_min_ord,
                                                    and_(result_set.c.id == result_set_min_ord.c.id,
                                                         result_set.c.ord == result_set_min_ord.c.min_ord))
                                              .filter(result_set.c.unit_id == self.unit.id).count())


#======================================== РАСЧЁТ ЗАЧИСЛЕННЫХ ДЕТЕЙ =============================#
class BaseEnrolledAlgorithm(BaseAlgorithm):

    __metaclass__ = ABCMeta

    def _get_fact_group(self):
        """
        Получаем фактические группы данного учреждения.
        """

        fact_group_this_unit = (self.session.query(AlchemyGroup).filter(AlchemyGroup.status == GroupStatusEnum.FACT,
                                                                        AlchemyGroup.unit_id == self.unit.id)
                                                                .options(load_only('id', 'work_type_id',
                                                                                   'is_family', 'type_id'))
                                                                .subquery())
        return fact_group_this_unit

    @abstractmethod
    def get_result_set(self):
        pass


class EnrolledCountByAge(BaseEnrolledAlgorithm):

    def get_result_set(self, **kwargs):

        return (self.session.query(AlchemyPupil).join(self._get_fact_group()).join(self._kids(**kwargs))
                                                .options(load_only('id'))
                                                .subquery())


class EnrolledCountByDisabled(BaseEnrolledAlgorithm):

    def get_result_set(self, **kwargs):

        fact_group_this_unit = self._get_fact_group()

        return (self.session.query(AlchemyPupil).join(fact_group_this_unit)
                                                .join(AlchemyGroupType,
                                                      and_(fact_group_this_unit.c.type_id ==
                                                           AlchemyGroupType.id,
                                                           or_(AlchemyGroupType.code == 'compensative',
                                                               AlchemyGroupType.code == 'health')))
                                                .join(self._kids(**kwargs))
                                                .options(load_only('id')).subquery())


class EnrolledCountByFamily(BaseEnrolledAlgorithm):

    def get_result_set(self, **kwargs):

        fact_group_this_unit = self._get_fact_group()

        return (self.session.query(AlchemyPupil).join(fact_group_this_unit)
                                                .join(self._kids(**kwargs))
                                                .filter(fact_group_this_unit.c.is_family)
                                                .options(load_only('id')).subquery())


class EnrolledCountByCorrect(BaseEnrolledAlgorithm):

    def get_result_set(self, **kwargs):

        fact_group_this_unit = self._get_fact_group()

        return (self.session.query(AlchemyPupil).join(fact_group_this_unit)
                                                .join(AlchemyGroupType,
                                                      and_(fact_group_this_unit.c.type_id == AlchemyGroupType.id,
                                                           AlchemyGroupType.code == 'compensative'))
                                                .join(self._kids(**kwargs))
                                                .options(load_only('id')).subquery())


class EnrolledCountByHealth(BaseEnrolledAlgorithm):

    def get_result_set(self, **kwargs):

        fact_group_this_unit = self._get_fact_group()

        return (self.session.query(AlchemyPupil).join(fact_group_this_unit)
                                                .join(AlchemyGroupType,
                                                      and_(fact_group_this_unit.c.type_id == AlchemyGroupType.id,
                                                           AlchemyGroupType.code == 'health'))
                                                .join(self._kids(**kwargs))
                                                .options(load_only('id')).subquery())


class EnrolledCountShortStay(BaseEnrolledAlgorithm):

    def get_result_set(self, **kwargs):

        fact_group_this_unit = self._get_fact_group()

        return (self.session.query(AlchemyPupil).join(fact_group_this_unit)
                                                .join(AlchemyGroupWorkType,
                                                      and_(fact_group_this_unit.c.work_type_id ==
                                                           AlchemyGroupWorkType.id,
                                                           AlchemyGroupWorkType.code == '3-5'))
                                                .join(self._kids(**kwargs))
                                                .options(load_only('id')).subquery())


class EnrolledCountWantChangeDou(BaseEnrolledAlgorithm):

    def get_result_set(self, **kwargs):

        fact_group_this_unit = self._get_fact_group()
        kids = self._kids(**kwargs)
        pupils = self.session.query(AlchemyPupil).join(fact_group_this_unit).subquery()
        pupils = (self.session.query(kids.c.id.label('children_id')).join(pupils, pupils.c.children_id == kids.c.id)
                                                                    .subquery())

        want_change_dou_status = self.session.query(AlchemyDeclarationStatus).filter(AlchemyDeclarationStatus.code ==
                                                                                     DSS.WANT_CHANGE_DOU).subquery()
        declarations_want_change_dou = (self.session.query(AlchemyDeclaration).join(want_change_dou_status)
                                                                              .join(pupils,
                                                                                    AlchemyDeclaration.children_id ==
                                                                                    pupils.c.children_id)
                                                                              .options(load_only('children_id'))
                                                                              .distinct()
                                                                              .subquery())
        return declarations_want_change_dou


class EnrolledIndex(BaseIndex):
    """
    Запросы группы "Количество детей зачисленных ... "
    """

    COUNT_BY_AGE = 1
    COUNT_DISABLED = 2
    COUNT_FAMILY = 3
    COUNT_CORRECT = 4
    COUNT_HEALTH = 5
    COUNT_SHORT_STAY = 6
    COUNT_WANT_CHANGE_DOU = 7

    ALGORITHM_MAP = {
        COUNT_BY_AGE: EnrolledCountByAge,
        COUNT_DISABLED: EnrolledCountByDisabled,
        COUNT_FAMILY: EnrolledCountByFamily,
        COUNT_CORRECT: EnrolledCountByCorrect,
        COUNT_HEALTH: EnrolledCountByHealth,
        COUNT_SHORT_STAY: EnrolledCountShortStay,
        COUNT_WANT_CHANGE_DOU: EnrolledCountWantChangeDou
    }

    def get_count(self, dou_type=GOVERMENT, index_type=COUNT_BY_AGE, **kwargs):

        if (not self.dou_type and dou_type != GOVERMENT) or (self.dou_type and self.dou_type.code != dou_type):
            return 0

        result_set = (EnrolledIndex.get_algorithm(index_type)(unit=self.unit, session=self.session)
                                   .get_result_set(**kwargs))

        return self.session.query(result_set).count()