# coding: utf-8
"""
Модуль функций подсчета показателей.
"""

from django.db.models import Q
import settings as gs_settings

from kinder.core.declaration.models import DeclarationUnit, Declaration
from kinder.core.declaration_status.models import DSS
from kinder.core.group.models import Pupil, GroupStatusEnum
from kinder.core.unit.models import UnitKind

from kinder.core.privilege.models import Privilege
from kinder.core.dict.models import HealthNeed
from kinder.core.alloc.helpers import GroupHelper

from constants import (AGE_CATEGORY_DELTAS, DELIVERY_TYPES, ALL,
                       APPLICATION_STATUSES, GROUP_TYPES,
                       AGE_SUBCATEGORY_MAP, AGE_CATEGORY_MAP, AGE_DELTAS,
                       GOVERMENT, get_report_period, get_last_day_of_academic_year)

# статусы заявлений для запросов
APPROVED_STATUS_CODES = [DSS.ZAGS_CHECKING, DSS.DUL_CONFIRMATING,
                         DSS.PRIV_CONFIRMATING, DSS.MED_CONFIRMATING,
                         DSS.TUTOR_CONFIRMATING, DSS.WANT_CHANGE_DOU,
                         DSS.REGISTERED]
REFUSED_STATUS_CODES = [DSS.REFUSED]
EXCLUDE_STATUSES = [DSS.ACCEPTED, DSS.ARCHIVE]

FEDERAL = 1
MUNICIPAL = 2
REGIONAL = 3
PRIVELEGE_TYPE = {
    FEDERAL: u'Федеральная',
    MUNICIPAL: u'Муниципальная',
    REGIONAL: u'Региональная'
}


def _get_base_queue_for_unit(calc, unit_id, dou_type, ages_const):

    result = []
    if unit_id in calc.queue_cache and dou_type in calc.queue_cache[unit_id]:
        result = calc.queue_cache[unit_id][dou_type][ages_const]

    return result


def count_of_children_in_short_stay(calc, unit_id, ages_const, dou_type=GOVERMENT):
    """
    Количество детей в очереди в группы кратковременного пребывания
    Считаем как заявки с типом work_type = кратковременный (3-5)(id=4)
    @param calc: экземпляр класса расчета
    @param ages_const: Константа границ возрастной категории
    @param unit_id: id учреждения, для которого считается очередь
    @return:
    """
    qs = _get_base_queue_for_unit(calc, unit_id, dou_type, ages_const)
    qs = filter(lambda x: x['work_type__code'] == '3-5', qs)
    return len(qs)


def count_of_children_in_queue_base(calc, unit_id, ages_const, dou_type=GOVERMENT):
    """
    Очередь как есть
    @param calc: экземпляр класса расчета
    @param ages_const: Константа границ возрастной категории
    @param unit_id: id учреждения, для которого считается очередь
    """
    qs = _get_base_queue_for_unit(calc, unit_id, dou_type, ages_const)
    return len(qs)


def count_of_children_in_privilege_queue(calc, unit_id, ages_const, privilege_type=None, dou_type=GOVERMENT):
    """
    Очередь только по льготникам
    отличается по списку статусов и наличию лгот
    @param calc: экземпляр класса расчета
    @param ages_const: Константа границ возрастной категории
    @param unit_id: id учреждения, для которого считается очередь
    @return:
    """
    qs = _get_base_queue_for_unit(calc, unit_id, dou_type, ages_const)
    qs = filter(lambda x: x['best_privilege__id'], qs)
    qs = filter(lambda x: x['status__code'] == DSS.REGISTERED, qs)
    qs = filter(lambda x: not x['desired_date'] or x['desired_date'] <= get_last_day_of_academic_year(), qs)
    if privilege_type:
        if privilege_type == REGIONAL:
            privilege_filter = Q(type=REGIONAL) | Q(type=MUNICIPAL)
        else:
            privilege_filter = Q(type=privilege_type)
        qs_pr = Privilege.objects.filter(privilege_filter).values_list('id', flat=True)
        qs = filter(lambda x: x['best_privilege__id'] in qs_pr, qs)
    return len(qs)


def children_without_place_count(calc, unit_id, ages_const, dou_type=GOVERMENT):
    """
    Подсчет количества детей, не обеспеченных местом в ДОУ
    @param calc: экземпляр класса расчета
    @param ages_const: Константа границ возрастной категории
    @param unit_id: id учреждения, для которого считается очередь
    """
    base_qs = _get_base_queue_for_unit(calc, unit_id, dou_type, ages_const)
    base_qs = filter(lambda x: not x['desired_date'] or x['desired_date'] <= get_last_day_of_academic_year(), base_qs)
    return len(base_qs)


def children_with_health_need_count(calc, unit_id, ages_const, comp=0, dou_type=GOVERMENT):
    """
    Подсчет количества детей  в очереди, имеющих потребности по здоровью
    @param calc: экземпляр класса расчета
    @param ages_const: Константа границ возрастной категории
    @param unit_id: id учреждения, для которого считается очередь
    @param comp: 0 все группы
    comp 1 компенсирующие группы
    comp 2 оздоровительные группы
    """
    qs = _get_base_queue_for_unit(calc, unit_id, dou_type, ages_const)
    # При подсчете детей в очереди с потребностью по здоровью учитывать только актуальный спрос.
    qs = filter(lambda x: not x['desired_date'] or x['desired_date'] <= get_last_day_of_academic_year(), qs)
    if comp == 0:
        qs_hn = HealthNeed.objects.exclude(name=u'Нет').values_list('id', flat=True)
    elif comp == 1:
        #id потребностей из документа расчета показателей
        qs_hn = range(1, 10) + [12, 13]
    elif comp == 2:
        qs_hn = [14, 15, 16]
    qs = filter(lambda x: x['children__health_need_id'] in qs_hn, qs)
    return len(qs)


def postponed_queue_children_count(calc, unit_id, ages_const, dou_type=GOVERMENT):
    """
    Количетство детей в отложенной очереди
    Заявки с желаемой датой зачисления больше 10.09.текущий_учебный_год
    + если комплектование прошло (дата окончания < текущей) то те дети у кого дата
    желаемого зачисления пустая либо меньше 31.12.текущий_год
    @param calc: экземпляр класса расчета
    @param ages_const: Константа границ возрастной категории
    @param unit_id: id учреждения, для которого считается очередь
    """
    base_qs = _get_base_queue_for_unit(calc, unit_id, dou_type, ages_const)

    base_qs = filter(lambda x: x['desired_date'] and x['desired_date'] > get_last_day_of_academic_year(), base_qs)
    count = len(base_qs)

    return count


#----------------------------------------------
# расчеты по количеству свободных мест
#----------------------------------------------
class GisdoGroupHelper(GroupHelper):
    """
    Отличается от родителя тем, что добавлены дополнительные фильтры
    """

    def getQuery(self, reporting=False):

        q, tmp_dict = super(GisdoGroupHelper, self).getQuery(reporting)

        # Группы кратковременного пребывания
        short_stay = getattr(self.context, 'short_stay', False)
        if short_stay:
            q = q.filter(work_type__code='3-5')

        # Потребность по здоровью
        health_need_filter = getattr(self.context, 'health_need_filter', False)
        if health_need_filter:
            # (#85387) Считать, как количество всех детей зачисленных в группы типа "Оздоровительные"
            # или "Компенсирующие" без учета наличия или отсутствия потребности по здоровью.
            q = q.filter(type__code__in=['health', 'compensative'])

        dou_type = getattr(self.context, 'dou_type', GOVERMENT)
        filter_by_dou_type = Q(unit__dou_type__code=dou_type)
        if dou_type == GOVERMENT: filter_by_dou_type |= Q(unit__dou_type=None)
        q = q.filter(filter_by_dou_type)

        return q, tmp_dict

    @classmethod
    def get_count(cls, context, age_category):

        total = 0
        for group in cls(context).getList():

            # Если у группы есть возрастная подкатегория, то соотношение проводится по ней
            if group.sub_age_cat:
                if group.sub_age_cat.code in AGE_SUBCATEGORY_MAP[age_category]:
                    total += group.cntFree
            # в противном случае по возрастной категории, которая обязательна.
            else:
                if group.age_cat.code in AGE_CATEGORY_MAP[age_category]:
                    total += group.cntFree

        return total


def get_capacities_count(unit, age_category=None, dou_type=GOVERMENT):
    """Подсчет показателей по свободным местам"""
    class Context(object):
        kind = 1  # магическое число указывает на то, что группа фактическая.
                  # использую потому, что это значение строго вшито в GroupHelper

        def __init__(self, unit_id):

            self.group_unit = unit_id
            self.dou_type = dou_type

    return GisdoGroupHelper.get_count(Context(unit.id), age_category)


def get_capacities_count_by_group_type(unit, group_type, age_category):
    """Подсчет показателей по свободным местам по типам групп"""
    class Context(object):
        kind = 1  # магическое число указывает на то, что группа фактическая

        def __init__(self, unit_id, short_stay=False, health_need_filter=False):

            self.group_unit = unit_id
            self.short_stay = short_stay
            self.health_need_filter = health_need_filter

    short_stay = health_need_filter = False
    if group_type == GROUP_TYPES['DISABLED']:
        health_need_filter = True
    if group_type == GROUP_TYPES['SHORT_STAY']:
        short_stay = True

    return GisdoGroupHelper.get_count(Context(unit.id, short_stay=short_stay, health_need_filter=health_need_filter)
                                                                            , age_category=age_category)